# -*- coding: utf-8 -*-
"""Python wrapper for command line
"""
import os
import subprocess
from biobb_common.tools import file_utils as fu
from typing import Optional, Union
import logging
from pathlib import Path


class CmdWrapper:
    """Command line wrapper using subprocess library
    """

    def __init__(self,
                 cmd: list[str],
                 shell_path: Union[str, Path] = os.getenv('SHELL', '/bin/sh'),
                 out_log: Optional[logging.Logger] = None,
                 err_log: Optional[logging.Logger] = None,
                 global_log: Optional[logging.Logger] = None,
                 env: Optional[dict] = None,
                 timeout: Optional[int] = None,
                 disable_logs: Optional[bool] = None) -> None:

        self.cmd = cmd
        self.shell_path = shell_path
        self.out_log = out_log
        self.err_log = err_log
        self.global_log = global_log
        self.env = env
        self.timeout = timeout
        self.disable_logs = disable_logs

    def log_output(self, exit_code: str, command: str, out: Optional[bytes] = None, err: Optional[bytes] = None, timeout: Optional[str] = None,
                   out_log: Optional[logging.Logger] = None, err_log: Optional[logging.Logger] = None, global_log: Optional[logging.Logger] = None) -> None:

        timeout_str = ''
        if timeout:
            timeout_str = f"Timeout: {timeout} seconds expired, killing process\n"
        command_str = f"Command '{command[0:80]}...' finalized with exit code {exit_code}"
        if out_log:
            out_log.info(command_str)
            if timeout_str:
                out_log.info(timeout_str)
            if out:
                out_log.info(out.decode("utf-8"))
        elif not self.disable_logs:
            print(command_str)
            if timeout_str:
                print(timeout_str)
            print("")
        if err_log and err:
            err_log.info(err.decode("utf-8"))

        if global_log:
            global_log.info(f"{fu.get_logs_prefix()}{command_str}")
            if timeout_str:
                global_log.info(f"{fu.get_logs_prefix()}{timeout_str}")

    def launch(self) -> int:
        cmd = " ".join(self.cmd)
        if self.out_log:
            self.out_log.info(f'Launching command (it may take a while): {cmd}')
        elif not self.disable_logs:
            print(f"\ncmd_wrapper command print: {cmd}")

        new_env = {**os.environ.copy(), **self.env} if self.env else os.environ.copy()
        process = subprocess.Popen(cmd,
                                   stdout=subprocess.PIPE,
                                   stderr=subprocess.PIPE,
                                   shell=True,
                                   executable=self.shell_path,
                                   env=new_env)
        try:
            out, err = process.communicate(timeout=self.timeout)
        except subprocess.TimeoutExpired:
            process.kill()
            out, err = process.communicate()
            process.returncode = 1
            self.log_output(exit_code=str(process.returncode), command=" ".join(self.cmd), out=out, err=err, timeout=str(self.timeout), out_log=self.out_log, err_log=self.err_log, global_log=self.global_log)
            return process.returncode

        process.wait()
        self.log_output(exit_code=str(process.returncode), command=" ".join(self.cmd), out=out, err=err, out_log=self.out_log, err_log=self.err_log, global_log=self.global_log)
        return process.returncode



name = "biobb_common"
__all__ = ["generic", "tools"]
__version__ = "5.2.1"


class BioBBGlobalProperties(dict):
    """Global properties container for all BiobbObject instances."""

    def __init__(self):
        super().__init__()

    def dict(self):
        """Create a shallow copy of the global properties."""
        return self.copy()

    def __repr__(self):
        """String representation."""
        return f"BioBBGlobalProperties({dict(self)})"


# Create a global instance
biobb_global_properties = BioBBGlobalProperties()



"""Tools to work with files
"""
import difflib
import functools
import logging
import os
import errno
import pathlib
import re
import shutil
import uuid
import warnings
import zipfile
from sys import platform
from pathlib import Path
import typing
from typing import Optional, Union
import sys
from contextlib import contextmanager


def create_unique_file_path(parent_dir: Optional[Union[str, Path]] = None, extension: Optional[Union[str, Path]] = None) -> str:
    if not parent_dir:
        parent_dir = Path.cwd()
    if not extension:
        extension = ""
    while True:
        name = f"{uuid.uuid4()}{extension}"
        file_path = Path.joinpath(Path(parent_dir).resolve(), name)
        if not file_path.exists():
            return str(file_path)


def create_dir(dir_path: str) -> str:
    """Returns the directory **dir_path** and create it if path does not exist.

    Args:
        dir_path (str): Path to the directory that will be created.

    Returns:
        str: Directory dir path.
    """
    if not Path(dir_path).exists():
        Path(dir_path).mkdir(exist_ok=True, parents=True)
    return str(Path(dir_path))


def create_stdin_file(intput_string: str) -> str:
    file_path = create_unique_file_path(extension=".stdin")
    with open(file_path, "w") as file_handler:
        file_handler.write(intput_string)
    return file_path


def create_unique_dir(
    path: str = "",
    prefix: str = "",
    number_attempts: int = 10,
    out_log: Optional[logging.Logger] = None,
) -> str:
    """Create a directory with a prefix + computed unique name. If the
    computed name collides with an existing file name it attemps
    **number_attempts** times to create another unique id and create
    the directory with the new name.

    Args:
        path (str): ('') Parent path of the new directory.
        prefix (str): ('') String to be added before the computed unique dir name.
        number_attempts (int): (10) number of times creating the directory if there's a name conflict.
        out_log (logger): (None) Python logger object.

    Returns:
        str: Directory dir path.
    """
    new_dir = prefix + str(uuid.uuid4())
    if path:
        new_dir = str(Path(path).joinpath(new_dir))
    for i in range(number_attempts):
        try:
            oldumask = os.umask(0)
            Path(new_dir).mkdir(mode=0o777, parents=True, exist_ok=False)
            if out_log:
                out_log.info("Directory successfully created: %s" % new_dir)
            os.umask(oldumask)
            return new_dir
        except OSError:
            if out_log:
                out_log.info(new_dir + " Already exists")
                out_log.info("Retrying %i times more" % (number_attempts - i))
            new_dir = prefix + str(uuid.uuid4().hex)
            if path:
                new_dir = str(Path(path).joinpath(new_dir))
            if out_log:
                out_log.info("Trying with: " + new_dir)
    raise FileExistsError


def get_working_dir_path(working_dir_path: Optional[Union[str, Path]] = None, restart: bool = False) -> str:
    """Return the directory **working_dir_path** and create it if working_dir_path
    does not exist. If **working_dir_path** exists a consecutive numerical suffix
    is added to the end of the **working_dir_path** and is returned.

    Args:
        working_dir_path (str): Path to the workflow results.
        restart (bool): If step result exists do not execute the step again.

    Returns:
        str: Path to the workflow results directory.
    """
    if not working_dir_path:
        return str(Path.cwd().resolve())

    working_dir_path = str(Path(working_dir_path).resolve())

    if (not Path(working_dir_path).exists()) or restart:
        return str(Path(working_dir_path))

    cont = 1
    while Path(str(working_dir_path)).exists():
        working_dir_path = (
            re.split(r"_[0-9]+$", str(working_dir_path))[0] + "_" + str(cont)
        )
        cont += 1
    return str(working_dir_path)


def zip_list(
    zip_file: Union[str, Path], file_list: typing.Sequence[Union[str, Path]], out_log: Optional[logging.Logger] = None
):
    """Compress all files listed in **file_list** into **zip_file** zip file.

    Args:
        zip_file (str): Output compressed zip file.
        file_list (:obj:`list` of :obj:`str`): Input list of files to be compressed.
        out_log (:obj:`logging.Logger`): Input log object.
    """
    file_list = list(file_list)
    file_list.sort()
    Path(zip_file).parent.mkdir(parents=True, exist_ok=True)
    with zipfile.ZipFile(zip_file, "w") as zip_f:
        inserted = []
        for index, f in enumerate(file_list):
            base_name = Path(f).name
            if base_name in inserted:
                base_name = "file_" + str(index) + "_" + base_name
            inserted.append(base_name)
            zip_f.write(f, arcname=base_name)
    if out_log:
        out_log.info("Adding:")
        # out_log.info(list(map(lambda x: str(Path(x).resolve().relative_to(Path.cwd())), file_list)))
        out_log.info(str(file_list))
        out_log.info("to: " + str(Path(zip_file).resolve()))


def unzip_list(
    zip_file: Union[str, Path], dest_dir: Optional[Union[str, Path]] = None, out_log: Optional[logging.Logger] = None
) -> list[str]:
    """Extract all files in the zipball file and return a list containing the
        absolute path of the extracted files.

    Args:
        zip_file (str): Input compressed zip file.
        dest_dir (str): Path to directory where the files will be extracted.
        out_log (:obj:`logging.Logger`): Input log object.

    Returns:
        :obj:`list` of :obj:`str`: list of paths of the extracted files.
    """
    with zipfile.ZipFile(zip_file, "r") as zip_f:
        zip_f.extractall(path=dest_dir)
        file_list = [str(Path(str(dest_dir)).joinpath(f)) for f in zip_f.namelist()]

    if out_log:
        out_log.info("Extracting: " + str(Path(zip_file).resolve()))
        out_log.info("to:")
        out_log.info(str(file_list))

    return file_list


def search_topology_files(
    top_file: Union[str, Path], out_log: Optional[logging.Logger] = None
) -> list[str]:
    """Search the top and itp files to create a list of the topology files

    Args:
        top_file (str): Topology GROMACS top file.
        out_log (:obj:`logging.Logger`): Input log object.

    Returns:
        :obj:`list` of :obj:`str`: list of paths of the extracted files.
    """
    top_dir_name = str(Path(top_file).parent)
    file_list = []
    pattern = re.compile(r"#include\s+\"(.+)\"")
    if Path(top_file).exists():
        with open(top_file) as tf:
            for line in tf:
                include_file = pattern.match(line.strip())
                if include_file:
                    found_file = str(Path(top_dir_name).joinpath(include_file.group(1)))
                    file_list += search_topology_files(found_file, out_log)
    else:
        if out_log:
            out_log.info("Ignored file %s" % top_file)
        return file_list
    return file_list + [str(top_file)]


def zip_top(
    zip_file: Union[str, Path],
    top_file: Union[str, Path],
    out_log: Optional[logging.Logger] = None,
    remove_original_files: bool = True,
) -> list[str]:
    """Compress all *.itp and *.top files in the cwd into **zip_file** zip file.

    Args:
        zip_file (str): Output compressed zip file.
        top_file (str): Topology TOP GROMACS file.
        out_log (:obj:`logging.Logger`): Input log object.

    Returns:
        :obj:`list` of :obj:`str`: list of compressed paths.
    """

    file_list = search_topology_files(top_file, out_log)
    zip_list(zip_file, file_list, out_log)
    if remove_original_files:
        rm_file_list(file_list, out_log)
    return file_list


def unzip_top(
    zip_file: Union[str, Path],
    out_log: Optional[logging.Logger] = None,
    unique_dir: Optional[Union[pathlib.Path, str]] = None,
) -> str:
    """Extract all files in the zip_file and copy the file extracted ".top" file to top_file.

    Args:
        zip_file (str): Input topology zipball file path.
        out_log (:obj:`logging.Logger`): Input log object.
        unique_dir (str): Directory where the topology will be extracted.

    Returns:
        str: Path to the extracted ".top" file.

    """
    unique_dir = unique_dir or create_unique_dir()
    top_list = unzip_list(zip_file, unique_dir, out_log)
    top_file = next(name for name in top_list if name.endswith(".top"))
    if out_log:
        out_log.info("Unzipping: ")
        out_log.info(zip_file)
        out_log.info("To: ")
        for file_name in top_list:
            out_log.info(file_name)
    return top_file


def get_logs_prefix():
    return 4 * " "


def create_incremental_name(path: Union[Path, str]) -> str:
    """Increment the name of the file by adding a number at the end.

    Args:
        path (str): path of the file.

    Returns:
        str: Incremented name of the file.
    """
    if (path_obj := Path(path)).exists():
        cont = 1
        while path_obj.exists():
            new_name = f'{path_obj.stem.rstrip("0123456789_")}_{cont}{path_obj.suffix}'
            path_obj = path_obj.with_name(new_name)
            cont += 1
    return str(path_obj)


def get_logs(
    path: Optional[Union[str, Path]] = None,
    prefix: Optional[str] = None,
    step: Optional[str] = None,
    can_write_console: bool = True,
    can_write_file: bool = True,
    out_log_path: Optional[Union[str, Path]] = None,
    err_log_path: Optional[Union[str, Path]] = None,
    level: str = "INFO",
    light_format: bool = False,
) -> tuple[logging.Logger, logging.Logger]:
    """Get the error and and out Python Logger objects.

    Args:
        path (str): (current working directory) Path to the log file directory.
        prefix (str): Prefix added to the name of the log file.
        step (str):  String added between the **prefix** arg and the name of the log file.
        can_write_console (bool): (True) If True, show log in the execution terminal.
        can_write_file (bool): (True) If True, write log to the log files.
        out_log_path (str): (None) Path to the out log file.
        err_log_path (str): (None) Path to the err log file.
        level (str): ('INFO') Set Logging level. ['CRITICAL','ERROR','WARNING','INFO','DEBUG','NOTSET']
        light_format (bool): (False) Minimalist log format.

    Returns:
        :obj:`tuple` of :obj:`logging.Logger` and :obj:`logging.Logger`: Out and err Logger objects.
    """
    out_log_path = out_log_path or "log.out"
    err_log_path = err_log_path or "log.err"
    # If paths are not absolute create and return them
    if not Path(out_log_path).is_absolute():
        out_log_path = create_incremental_name(create_name(path=path, prefix=prefix, step=step, name=str(out_log_path)))
    if not Path(err_log_path).is_absolute():
        err_log_path = create_incremental_name(create_name(path=path, prefix=prefix, step=step, name=str(err_log_path)))
    # Create logging objects
    out_Logger = logging.getLogger(str(out_log_path))
    err_Logger = logging.getLogger(str(err_log_path))

    # Create logging format
    logFormatter = logging.Formatter(
        "%(asctime)s [%(threadName)-12.12s] [%(levelname)-5.5s]  %(message)s"
    )
    if light_format:
        logFormatter = logging.Formatter("%(asctime)s %(message)s", "%H:%M:%S")

    if can_write_file:
        prefix = prefix if prefix else ""
        step = step if step else ""
        path = path if path else str(Path.cwd())

        # Create dir if it not exists
        create_dir(str(Path(out_log_path).resolve().parent))

        # Create FileHandler
        out_fileHandler = logging.FileHandler(out_log_path, mode="a", encoding=None, delay=True)
        err_fileHandler = logging.FileHandler(err_log_path, mode="a", encoding=None, delay=True)
        # Asign format to FileHandler
        out_fileHandler.setFormatter(logFormatter)
        err_fileHandler.setFormatter(logFormatter)

        # Assign FileHandler to logging object
        if not len(out_Logger.handlers):
            out_Logger.addHandler(out_fileHandler)
            err_Logger.addHandler(err_fileHandler)

    if can_write_console:
        console_out = logging.StreamHandler(stream=sys.stdout)
        console_err = logging.StreamHandler(stream=sys.stderr)
        console_out.setFormatter(logFormatter)
        console_err.setFormatter(logFormatter)
        # Assign consoleHandler to logging objects as aditional output
        if len(out_Logger.handlers) < 2:
            out_Logger.addHandler(console_out)
            err_Logger.addHandler(console_err)

    # Set logging level level
    out_Logger.setLevel(level)
    err_Logger.setLevel(level)

    return out_Logger, err_Logger


def launchlogger(func):
    """Decorator to create the out_log and err_log"""
    @functools.wraps(func)
    def wrapper_log(*args, **kwargs):
        create_dir(create_name(path=args[0].path))
        if args[0].disable_logs:
            return func(*args, **kwargs)

        # Create local out_log and err_log
        args[0].out_log, args[0].err_log = get_logs(
            path=args[0].path,
            prefix=args[0].prefix,
            step=args[0].step,
            can_write_console=args[0].can_write_console_log,
            can_write_file=args[0].can_write_file_log,
            out_log_path=args[0].out_log_path,
            err_log_path=args[0].err_log_path
        )

        # Run the function and capture its return value
        value = func(*args, **kwargs)

        # Close and remove handlers from out_log and err_log
        for log in [args[0].out_log, args[0].err_log]:
            # Create a copy [:] of the handler list to be able to modify it while we are iterating
            handlers = log.handlers[:]
            for handler in handlers:
                handler.close()
                log.removeHandler(handler)

        return value

    return wrapper_log


def log(string: str, local_log: Optional[logging.Logger] = None, global_log: Optional[logging.Logger] = None):
    """Checks if log exists

    Args:
        string (str): Message to log.
        local_log (:obj:`logging.Logger`): local log object.
        global_log (:obj:`logging.Logger`): global log object.

    """
    if local_log:
        local_log.info(string)
    if global_log:
        global_log.info(get_logs_prefix() + string)


def human_readable_time(time_ps: int) -> str:
    """Transform **time_ps** to a human readable string.

    Args:
        time_ps (int): Time in pico seconds.

    Returns:
        str: Human readable time.
    """
    time_units = [
        "femto seconds",
        "pico seconds",
        "nano seconds",
        "micro seconds",
        "mili seconds",
    ]
    t = time_ps * 1000
    for tu in time_units:
        if t < 1000:
            return str(t) + " " + tu

        t = int(t/1000)
    return str(time_ps)


def check_properties(obj: object, properties: dict, reserved_properties: Optional[list[str]] = None):
    if not reserved_properties:
        reserved_properties = []
    error_properties = set(
        [prop for prop in properties.keys() if prop not in obj.__dict__.keys()]
    )
    error_properties -= set(["system", "working_dir_path"] + list(reserved_properties))
    for error_property in error_properties:
        close_property_list = difflib.get_close_matches(
            error_property, obj.__dict__.keys(), n=1, cutoff=0.01
        )
        close_property = close_property_list[0] if close_property_list else ""
        warnings.warn(
            "Warning: %s is not a recognized property. The most similar property is: %s"
            % (error_property, close_property)
        )


def create_name(
    path: Optional[Union[str, Path]] = None, prefix: Optional[str] = None,
    step: Optional[str] = None, name: Optional[str] = None
) -> str:
    """Return file name.

    Args:
        path (str): Path to the file directory.
        prefix (str): Prefix added to the name of the file.
        step (str):  String added between the **prefix** arg and the **name** arg of the file.
        name (str): Name of the file.

    Returns:
        str: Composed file name.
    """
    name = "" if name is None else name.strip()
    if step:
        if name:
            name = step + "_" + name
        else:
            name = step
    if prefix:
        prefix = prefix.replace("/", "_")
        if name:
            name = prefix + "_" + name
        else:
            name = prefix
    if path:
        if name:
            name = str(Path(path).joinpath(name))
        else:
            name = str(path)
    return name


def write_failed_output(file_name: str):
    with open(file_name, "w") as f:
        f.write("Error\n")


def rm(file_name: Union[str, Path]) -> Optional[Union[str, Path]]:
    try:
        file_path = pathlib.Path(file_name)
        if file_path.exists():
            if file_path.is_dir():
                shutil.rmtree(file_name)
                return file_name
            if file_path.is_file():
                Path(file_name).unlink()
                return file_name
    except Exception:
        pass
    return None


def rm_file_list(
    file_list: typing.Sequence[Union[str, Path]], out_log: Optional[logging.Logger] = None
) -> list[str]:
    removed_files = [str(f) for f in file_list if rm(f)]
    if len(removed_files) > 0 and out_log:
        log("Removed: %s" % str(removed_files), out_log)
    return removed_files


def check_complete_files(output_file_list: list[Union[str, Path]]) -> bool:
    for output_file in filter(None, output_file_list):
        output_file = Path(str(output_file))
        file_exists = output_file.is_file() and output_file.stat().st_size > 0
        dir_exists = output_file.is_dir() and any(output_file.iterdir())
        if not file_exists and not dir_exists:
            return False
    return True


def copytree_new_files_only(source, destination):
    """
    Recursively copies files from source to destination only if they don't
    already exist in the destination.
    """
    if not os.path.exists(destination):
        os.makedirs(destination)

    for dirpath, dirnames, filenames in os.walk(source):
        # Create a corresponding directory in the destination
        relative_path = os.path.relpath(dirpath, source)
        dest_dir = os.path.join(destination, relative_path)
        if not os.path.exists(dest_dir):
            os.makedirs(dest_dir)

        # Copy files that do not exist or have newer modification times
        for filename in filenames:
            src_file_path = os.path.join(dirpath, filename)
            dest_file_path = os.path.join(dest_dir, filename)

            if not os.path.exists(dest_file_path) or os.path.getmtime(src_file_path) > os.path.getmtime(dest_file_path):
                shutil.copy2(src_file_path, dest_file_path)


def copy_to_container(container_path: Optional[Union[str, Path]], container_volume_path: str,
                      io_dict: dict, out_log: Optional[logging.Logger] = None) -> dict:
    if not container_path:
        return io_dict

    unique_dir = str(Path(create_unique_dir()).resolve())
    container_io_dict: dict = {"in": {}, "out": {}, "unique_dir": unique_dir}

    # IN files COPY and assign INTERNAL PATH
    for file_ref, file_path in io_dict["in"].items():
        if file_path:
            if Path(file_path).exists():
                shutil.copy2(file_path, unique_dir)
                log(f"Copy: {file_path} to {unique_dir}")
                container_io_dict["in"][file_ref] = str(
                    Path(container_volume_path).joinpath(Path(file_path).name)
                )
            else:
                # Default files in GMXLIB path like gmx_solvate -> input_solvent_gro_path (spc216.gro)
                container_io_dict["in"][file_ref] = file_path

    # OUT files assign INTERNAL PATH
    for file_ref, file_path in io_dict["out"].items():
        if file_path:
            container_io_dict["out"][file_ref] = str(
                Path(container_volume_path).joinpath(Path(file_path).name)
            )

    return container_io_dict


def copy_to_host(container_path: str, container_io_dict: dict, io_dict: dict):
    if not container_path:
        return

    # OUT files COPY
    for file_ref, file_path in container_io_dict["out"].items():
        if file_path:
            container_file_path = str(
                Path(container_io_dict["unique_dir"]).joinpath(Path(file_path).name)
            )
            if Path(container_file_path).exists():
                shutil.copy2(container_file_path, io_dict["out"][file_ref])


def create_cmd_line(
    cmd: list[str],
    container_path: Optional[Union[str, Path]] = "",
    host_volume: Optional[Union[str, Path]] = None,
    container_volume: Optional[Union[str, Path]] = None,
    container_working_dir: Optional[Union[str, Path]] = None,
    container_user_uid: Optional[str] = None,
    container_shell_path: Optional[Union[str, Path]] = None,
    container_image: Optional[Union[str, Path]] = None,
    out_log: Optional[logging.Logger] = None,
    global_log: Optional[logging.Logger] = None
) -> list[str]:
    container_path = container_path or ""
    if str(container_path).endswith("singularity"):
        log("Using Singularity image %s" % container_image, out_log, global_log)
        if not Path(str(container_image)).exists():
            log(
                f"{container_image} does not exist trying to pull it",
                out_log,
                global_log,
            )
            container_image_name = str(Path(str(container_image)).with_suffix(".sif").name)
            singularity_pull_cmd = [
                str(container_path),
                "pull",
                "--name",
                str(container_image_name),
                str(container_image),
            ]
            try:
                from biobb_common.command_wrapper import cmd_wrapper

                cmd_wrapper.CmdWrapper(cmd=singularity_pull_cmd, out_log=out_log).launch()
                if Path(container_image_name).exists():
                    container_image = container_image_name
                else:
                    raise FileNotFoundError
            except FileNotFoundError:
                log(f"{' '.join(singularity_pull_cmd)} not found", out_log, global_log)
                raise FileNotFoundError
        singularity_cmd: list[str] = [
            str(container_path),
            "exec",
            "-e",
            "--bind",
            str(host_volume) + ":" + str(container_volume),
            str(container_image),
        ]
        # If we are working on a mac remove -e option because is still no available
        if platform == "darwin":
            if "-e" in singularity_cmd:
                singularity_cmd.remove("-e")

        cmd = ['"' + " ".join(cmd) + '"']
        singularity_cmd.extend([str(container_shell_path), "-c"])
        return singularity_cmd + cmd

    elif str(container_path).endswith("docker"):
        log("Using Docker image %s" % container_image, out_log, global_log)
        docker_cmd = [str(container_path), "run"]
        if container_working_dir:
            docker_cmd.append("-w")
            docker_cmd.append(str(container_working_dir))
        if container_volume:
            docker_cmd.append("-v")
            docker_cmd.append(str(host_volume) + ":" + str(container_volume))
        if container_user_uid:
            docker_cmd.append("--user")
            docker_cmd.append(container_user_uid)

        docker_cmd.append(str(container_image))

        cmd = ['"' + " ".join(cmd) + '"']
        docker_cmd.extend([str(container_shell_path), "-c"])
        return docker_cmd + cmd

    elif str(container_path).endswith("pcocc"):
        # pcocc run -I racov56:pmx cli.py mutate -h
        log("Using pcocc image %s" % container_image, out_log, global_log)
        pcocc_cmd = [str(container_path), "run", "-I", str(container_image)]
        if container_working_dir:
            pcocc_cmd.append("--cwd")
            pcocc_cmd.append(str(container_working_dir))
        if container_volume:
            pcocc_cmd.append("--mount")
            pcocc_cmd.append(str(host_volume) + ":" + str(container_volume))
        if container_user_uid:
            pcocc_cmd.append("--user")
            pcocc_cmd.append(container_user_uid)

        cmd = ['\\"' + " ".join(cmd) + '\\"']
        pcocc_cmd.extend([str(container_shell_path), "-c"])
        return pcocc_cmd + cmd

    else:
        # log('Not using any container', out_log, global_log)
        return cmd


def get_doc_dicts(doc: Optional[str]):
    regex_argument = re.compile(
        r"(?P<argument>\w*)\ *(?:\()(?P<type>\w*)(?:\)):?\ *(?P<optional>\(\w*\):)?\ *(?P<description>.*?)(?:\.)\ *(?:File type:\ *)(?P<input_output>\w+)\.\ *(\`(?:.+)\<(?P<sample_file>.*?)\>\`\_\.)?\ *(?:Accepted formats:\ *)(?P<formats>.+)(?:\.)?"
    )
    regex_argument_formats = re.compile(
        r"(?P<extension>\w*)\ *(\(\ *)\ *edam\ *:\ *(?P<edam>\w*)"
    )
    regex_property = re.compile(
        r"(?:\*\ *\*\*)(?P<property>.*?)(?:\*\*)\ *(?:\(\*)(?P<type>\w*)(?:\*\))\ *\-\ ?(?:\()(?P<default_value>.*?)(?:\))\ *(?:(?:\[)(?P<wf_property>WF property)(?:\]))?\ *(?:(?:\[)(?P<range_start>[\-]?\d+(?:\.\d+)?)\~(?P<range_stop>[\-]?\d+(?:\.\d+)?)(?:\|)?(?P<range_step>\d+(?:\.\d+)?)?(?:\]))?\ *(?:(?:\[)(.*?)(?:\]))?\ *(?P<description>.*)"
    )
    regex_property_value = re.compile(
        r"(?P<value>\w*)\ *(?:(?:\()(?P<description>.*?)?(?:\)))?"
    )

    doc_lines = list(
        map(str.strip, filter(lambda line: line.strip(), str(doc).splitlines()))
    )
    args_index = doc_lines.index(
        next(filter(lambda line: line.lower().startswith("args"), doc_lines))
    )
    properties_index = doc_lines.index(
        next(filter(lambda line: line.lower().startswith("properties"), doc_lines))
    )
    examples_index = doc_lines.index(
        next(filter(lambda line: line.lower().startswith("examples"), doc_lines))
    )
    arguments_lines_list = doc_lines[args_index + 1: properties_index]
    properties_lines_list = doc_lines[properties_index + 1: examples_index]

    doc_arguments_dict = {}
    for argument_line in arguments_lines_list:
        match_argument = regex_argument.match(argument_line)
        argument_dict = match_argument.groupdict() if match_argument is not None else {}
        argument_dict["formats"] = {
            match.group("extension"): match.group("edam")
            for match in regex_argument_formats.finditer(argument_dict["formats"])
        }
        doc_arguments_dict[argument_dict.pop("argument")] = argument_dict

    doc_properties_dict = {}
    for property_line in properties_lines_list:
        match_property = regex_property.match(property_line)
        property_dict = match_property.groupdict() if match_property is not None else {}
        property_dict["values"] = None
        if "Values:" in property_dict["description"]:
            property_dict["description"], property_dict["values"] = property_dict[
                "description"
            ].split("Values:")
            property_dict["values"] = {
                match.group("value"): match.group("description")
                for match in regex_property_value.finditer(property_dict["values"])
                if match.group("value")
            }
        doc_properties_dict[property_dict.pop("property")] = property_dict

    return doc_arguments_dict, doc_properties_dict


def check_argument(
    path: Optional[pathlib.Path],
    argument: str,
    optional: bool,
    module_name: str,
    input_output: Optional[str] = None,
    output_files_created: bool = False,
    type: Optional[str] = None,
    extension_list: Optional[list[str]] = None,
    raise_exception: bool = True,
    check_extensions: bool = True,
    out_log: Optional[logging.Logger] = None,
) -> None:
    if optional and not path:
        return None

    if input_output in ["in", "input"]:
        input_file = True
    elif input_output in ["out", "output"]:
        input_file = False
    else:
        unable_to_determine_string = (
            f"{module_name} {argument}: Unable to determine if input or output file."
        )
        log(unable_to_determine_string, out_log)
        if raise_exception:
            raise FileNotFoundError(
                errno.ENOENT, os.strerror(errno.ENOENT), unable_to_determine_string
            )
        warnings.warn(unable_to_determine_string)

    if input_file or output_files_created:
        not_found_error_string = (
            f"Path {path} --- {module_name}: Unexisting {argument} file."
        )
        if not Path(str(path)).exists():
            log(not_found_error_string, out_log)
            if raise_exception:
                raise FileNotFoundError(
                    errno.ENOENT, os.strerror(errno.ENOENT), not_found_error_string
                )
            warnings.warn(not_found_error_string)
    # else:
    #     if not path.parent.exists():
    #         not_found_dir_error_string = f"Path {path.parent} --- {module_name}: Unexisting {argument} directory."
    #         log(not_found_dir_error_string, out_log)
    #         if raise_exception:
    #             raise FileNotFoundError(errno.ENOENT, os.strerror(errno.ENOENT), not_found_dir_error_string)
    #         warnings.warn(not_found_dir_error_string)

    if check_extensions and extension_list and type != "dir":
        no_extension_error_string = f"{module_name} {argument}: {path} has no extension. If you want to suppress this message, please set the check_extensions property to False"
        if not Path(str(path)).suffix:
            log(no_extension_error_string)
            warnings.warn(no_extension_error_string)
        else:
            not_valid_extension_error_string = f"{module_name} {argument}: {path} extension is not in the valid extensions list: {extension_list}. If you want to suppress this message, please set the check_extensions property to False"
            if not Path(str(path)).suffix[1:].lower() in extension_list:
                log(not_valid_extension_error_string)
                warnings.warn(not_valid_extension_error_string)


@contextmanager
def change_dir(destination):
    """Context manager for changing directory."""
    cwd = os.getcwd()
    if not Path(destination).exists():
        os.makedirs(destination)
    try:
        os.chdir(destination)
        yield
    finally:
        os.chdir(cwd)



"""Module containing the BiobbObject generic parent class."""
import difflib
import importlib
import os
import shutil
import warnings
import argparse
from logging import Logger
from pathlib import Path
from pydoc import locate
from sys import platform
from typing import Optional, Union
from biobb_common.configuration import settings
from biobb_common.command_wrapper import cmd_wrapper
from biobb_common.tools import file_utils as fu
from biobb_common import biobb_global_properties


class BiobbObject:
    """
    | biobb_common BiobbObject
    | Generic parent class for the rest of the Biobb clases.
    | The BiobbOject class contains all the properties and methods that are common to all the biobb blocks.

    Args:
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):

            * **io_dict** (*dict*) - ({}) Input/Output files dictionary.
            * **container_path** (*str*) - (None)  Path to the binary executable of your container.
            * **container_image** (*str*) - (None) Container Image identifier.
            * **container_volume_path** (*str*) - ("/data") Path to an internal directory in the container.
            * **container_working_dir** (*str*) - (None) Path to the internal CWD in the container.
            * **container_user_id** (*str*) - (None) User number id to be mapped inside the container.
            * **container_shell_path** (*str*) - ("/bin/bash -c") Path to the binary executable of the container shell.
            * **container_generic_command** (*str*) - ("run") Which command typically run or exec will be used to execute your image.
            * **stage_io_dict** (*dict*) - ({}) Stage Input/Output files dictionary.
            * **sandbox_path** (*str*) - ("./") [WF property] Parent path to the sandbox directory.
            * **disable_sandbox** (*bool*) - (False) Disable the use of temporal unique directories aka sandbox. Only for local execution.
            * **global_properties_list** (*list*) - ([]) list of global properties.
            * **chdir_sandbox** (*bool*) - (False) Change directory to the sandbox using just file names in the command line. Only for local execution.
            * **binary_path** (*str*) - ('') Path to the binary executable.
            * **disable_logs** (*bool*) - (False) Disable the logs.
            * **global_log** (*Logger object*) - (None) Log from the main workflow.
            * **out_log** (*Logger object*) - (None) Log from the step.
            * **err_log** (*Logger object*) - (None) Error log from the step.
            * **out_log_path** (*str*) - (None) Path to the log file.
            * **err_log_path** (*str*) - (None) Path to the error log file.
            * **can_write_console_log** (*bool*) - (True) Can write console log.
            * **can_write_file_log** (*bool*) - (True) Can write file log.
            * **prefix** (*str*) - (None) Prefix if provided.
            * **step** (*str*) - (None) Name of the step.
            * **path** (*str*) - ('') Absolute path to the step working dir.
            * **remove_tmp** (*bool*) - (True) [WF property] Remove temporal files.
            * **restart** (*bool*) - (False) [WF property] Do not execute if output files exist.
            * **cmd** (*list*) - ([]) Command line list, NOT read from the dictionary.
            * **return_code** (*int*) - (0) Return code of the command execution, NOT read from the dictionary.
            * **timeout** (*int*) - (None) Timeout for the execution of the command.
            * **tmp_files** (*list*) - ([]) list of temporal files, NOT read from the dictionary.
            * **env_vars_dict** (*dict*) - ({}) Environment Variables dictionary.
            * **shell_path** (*str*) - ("/bin/bash") Path to the binary executable of the shell.
            * **dev** (*str*) - (None) Development options.
            * **check_extensions** (*bool*) - (True) Check extensions of the input/output files.
            * **check_var_typing** (*bool*) - (True) Check typing of the input/output files.
            * **locals_var_dict** (*dict*) - ({}) Local variables dictionary.
            * **doc_arguments_dict** (*dict*) - ({}) Documentation arguments dictionary.
            * **doc_properties_dict** (*dict*) - ({}) Documentation properties dictionary.


    """

    def __init__(self, properties=None, **kwargs) -> None:  # type: ignore
        # Merge global properties, priorizating local ones
        properties = biobb_global_properties.dict() | properties or {}

        # Input/Output files
        self.io_dict: dict[str, dict[str, str]] = {"in": {}, "out": {}}

        # container Specific
        self.container_path: Optional[str] = properties.get("container_path")
        self.container_image: str = properties.get("container_image", '')
        self.container_volume_path: str = properties.get("container_volume_path", "/data")
        self.container_working_dir: Optional[str] = properties.get("container_working_dir")
        self.container_user_id: Optional[str] = properties.get("container_user_id")
        self.container_shell_path: str = properties.get("container_shell_path", "/bin/bash -c")
        self.container_generic_command: str = properties.get("container_generic_command", "run")

        # stage
        self.stage_io_dict: dict[str, dict[str, str]] = {"in": {}, "out": {}}
        self.sandbox_path: Union[str, Path] = properties.get("sandbox_path", Path().cwd())
        self.disable_sandbox: bool = properties.get("disable_sandbox", False)

        # Properties common in all BB
        self.global_properties_list: list[str] = properties.get("global_properties_list", [])
        self.chdir_sandbox: bool = properties.get("chdir_sandbox", False)
        self.binary_path: str = properties.get("binary_path", '')
        self.disable_logs: bool = properties.get("disable_logs", False)
        self.global_log: Optional[Logger] = properties.get("global_log", None)
        self.out_log: Optional[Logger] = None
        self.err_log: Optional[Logger] = None
        self.out_log_path: Optional[Union[Path, str]] = properties.get("out_log_path", None)
        self.err_log_path: Optional[Union[Path, str]] = properties.get("err_log_path", None)
        self.can_write_console_log: bool = properties.get("can_write_console_log", True)
        self.can_write_file_log: bool = properties.get("can_write_file_log", True)
        self.prefix: Optional[str] = properties.get("prefix", None)
        self.step: Optional[str] = properties.get("step", None)
        self.path: str = properties.get("path", "")
        self.remove_tmp: bool = properties.get("remove_tmp", True)
        self.restart: bool = properties.get("restart", False)
        self.cmd: list[str] = []
        self.return_code: int = 0
        self.timeout: Optional[int] = properties.get("timeout", None)
        self.tmp_files: list[Union[str, Path]] = []
        self.env_vars_dict: dict = properties.get("env_vars_dict", {})
        self.shell_path: Union[str, Path] = properties.get("shell_path", os.getenv("SHELL", "/bin/bash"))
        self.dev: Optional[str] = properties.get("dev", None)
        self.check_extensions: bool = properties.get("check_extensions", True)
        self.check_var_typing: bool = properties.get("check_var_typing", True)
        self.locals_var_dict: dict[str, str] = dict()
        self.doc_arguments_dict, self.doc_properties_dict = fu.get_doc_dicts(self.__doc__)

        try:
            self.version = importlib.import_module(
                self.__module__.split(".")[0]
            ).__version__
        except Exception:
            self.version = None

    def check_arguments(
            self,
            output_files_created: bool = False,
            raise_exception: bool = True
    ):
        for argument, argument_dict in self.doc_arguments_dict.items():
            fu.check_argument(
                path=Path(self.locals_var_dict[argument])
                if self.locals_var_dict.get(argument)
                else None,
                argument=argument,
                optional=argument_dict.get("optional", False),
                module_name=self.__module__,
                input_output=argument_dict.get(
                    "input_output", "").lower().strip(),
                output_files_created=output_files_created,
                type=argument_dict.get("type", None),
                extension_list=list(argument_dict.get("formats")),
                check_extensions=self.check_extensions,
                raise_exception=raise_exception,
                out_log=self.out_log,
            )
        if output_files_created:
            fu.log("", self.out_log, self.global_log)

    def check_properties(
        self,
        properties: dict,
        reserved_properties: Optional[set[str]] = None,
        check_var_typing: bool = False,
    ):
        if not reserved_properties:
            reserved_properties = set()
        reserved_properties = {"system", "working_dir_path", "tool"}.union(reserved_properties)
        reserved_properties = reserved_properties.union(set(self.global_properties_list))
        error_properties = set([prop for prop in properties.keys() if prop not in self.__dict__.keys()])

        # Check types
        if check_var_typing and self.doc_properties_dict:
            for prop, value in properties.items():
                if self.doc_properties_dict.get(prop):
                    property_type = self.doc_properties_dict[prop].get("type")
                    classinfo: object = locate(property_type).__class__
                    if classinfo == type:
                        classinfo = locate(property_type)
                    if not isinstance(value, classinfo):  # type: ignore
                        warnings.warn(
                            f"Warning: {prop} property type not recognized. Got {type(value)} Expected {locate(property_type)}"
                        )

        error_properties = set(
            [prop for prop in properties.keys() if prop not in self.__dict__.keys()]
        )
        error_properties -= reserved_properties
        for error_property in error_properties:
            close_property = difflib.get_close_matches(
                error_property, self.__dict__.keys(), n=1, cutoff=0.01
            )
            close_property = close_property[0] if close_property else ""  # type: ignore
            warnings.warn(
                "Warning: %s is not a recognized property. The most similar property is: %s"
                % (error_property, close_property)
            )

    def check_init(self, properties):
        """Check that the arguments and properties passed have the correct types and formats."""
        self.check_properties(properties)
        self.check_arguments()

    def check_restart(self) -> bool:
        if self.version:
            fu.log(
                f"Module: {self.__module__} Version: {self.version}",
                self.out_log, self.global_log
            )

        if self.restart:
            if fu.check_complete_files(self.io_dict["out"].values()):  # type: ignore
                fu.log("Restart is enabled, this step: %s will the skipped" % self.step, self.out_log, self.global_log)
                return True
        return False

    def stage_files(self):
        """Stage the input/output files in a temporal unique directory aka sandbox."""
        if self.disable_sandbox:
            self.stage_io_dict = self.io_dict.copy()
            # If we are not using a sandbox, we use the current working directory as the unique directory
            self.stage_io_dict["unique_dir"] = os.getcwd()
            return
        # Create a unique directory for the sandbox
        unique_dir = str(Path(fu.create_unique_dir(path=str(self.sandbox_path), prefix="sandbox_", out_log=self.out_log)).resolve())
        self.stage_io_dict = {"in": {}, "out": {}, "unique_dir": unique_dir}

        # Only remove unique_dir if using sandbox
        self.tmp_files.append(unique_dir)

        for io in ["in", "out"]:
            for file_ref, file_path in self.io_dict.get(io, {}).items():
                if not file_path:
                    # Skip optional files not set
                    continue
                file_path = Path(file_path)
                # Assign INTERNAL PATH to IN/OUT files
                if file_path.exists() or io == "out":
                    if io == "in":
                        fu.log(f"Copy to stage: {file_path} --> {unique_dir.split('/')[-1]}", self.out_log)
                        doc = self.doc_arguments_dict.get(file_ref)
                        if doc and doc['type'] == 'dir' and file_path.suffix != '.zip':
                            shutil.copytree(file_path, os.path.join(unique_dir, file_path.name))
                        else:
                            shutil.copy2(file_path, unique_dir)
                    # Container
                    if self.container_path:
                        self.stage_io_dict[io][file_ref] = os.path.join(self.container_volume_path, file_path.name)
                    # Local
                    else:
                        self.stage_io_dict[io][file_ref] = os.path.join(unique_dir, file_path.name)
                        if self.chdir_sandbox:
                            self.stage_io_dict[io][file_ref] = file_path.name
                elif io == "in":
                    # Default IN files in GMXLIB path like gmx_solvate -> input_solvent_gro_path (spc216.gro)
                    self.stage_io_dict[io][file_ref] = file_path.name

    def create_cmd_line(self) -> None:
        """ The method modifies the `self.cmd` attribute in-place to contain the final
        command line that will be executed based on the container type. """
        # Not documented and not listed option, only for devs
        if self.dev:
            fu.log(f"Adding development options: {self.dev}",
                   self.out_log, self.global_log)
            self.cmd += self.dev.split()

        # Containers
        host_volume: str = str(self.stage_io_dict.get("unique_dir", ''))
        self.container_path = self.container_path or ""
        # Singularity
        if self.container_path.endswith("singularity"):
            fu.log(
                "Using Singularity image %s" % self.container_image,
                self.out_log,
                self.global_log,
            )
            if not Path(self.container_image).exists():
                fu.log(
                    f"{self.container_image} does not exist trying to pull it",
                    self.out_log,
                    self.global_log,
                )
                container_image_name = str(
                    Path(self.container_image).with_suffix(".sif").name
                )
                singularity_pull_cmd = [
                    self.container_path,
                    "pull",
                    "--name",
                    container_image_name,
                    self.container_image,
                ]
                try:
                    from biobb_common.command_wrapper import cmd_wrapper

                    cmd_wrapper.CmdWrapper(
                        singularity_pull_cmd, self.shell_path, self.out_log
                    ).launch()
                    if Path(container_image_name).exists():
                        self.container_image = container_image_name
                    else:
                        raise FileNotFoundError
                except FileNotFoundError:
                    fu.log(
                        f"{' '.join(singularity_pull_cmd)} not found",
                        self.out_log,
                        self.global_log,
                    )
                    raise FileNotFoundError
            singularity_cmd = [
                self.container_path,
                self.container_generic_command,
                "-e",
            ]

            if self.env_vars_dict:
                singularity_cmd.append("--env")
                singularity_cmd.append(
                    ",".join(
                        f"{env_var_name}='{env_var_value}'"
                        for env_var_name, env_var_value in self.env_vars_dict.items()
                    )
                )

            singularity_cmd.extend(
                [
                    "--bind",
                    host_volume + ":" + self.container_volume_path,
                    self.container_image,
                ]
            )

            # If we are working on a mac remove -e option because is still no available
            if platform == "darwin":
                if "-e" in singularity_cmd:
                    singularity_cmd.remove("-e")

            if not self.cmd and not self.container_shell_path:
                fu.log(
                    "WARNING: The command-line is empty your container should know what to do automatically.",
                    self.out_log,
                    self.global_log,
                )
            else:
                cmd = ['"' + " ".join(self.cmd) + '"']
                singularity_cmd.append(self.container_shell_path)
                singularity_cmd.extend(cmd)
            self.cmd = singularity_cmd
        # Docker
        elif self.container_path.endswith("docker"):
            fu.log("Using Docker image %s" % self.container_image,
                   self.out_log, self.global_log)
            docker_cmd = [self.container_path, self.container_generic_command]
            if self.env_vars_dict:
                for env_var_name, env_var_value in self.env_vars_dict.items():
                    docker_cmd.append("-e")
                    docker_cmd.append(f"{env_var_name}='{env_var_value}'")
            if self.container_working_dir:
                docker_cmd.append("-w")
                docker_cmd.append(self.container_working_dir)
            if self.container_volume_path:
                docker_cmd.append("-v")
                docker_cmd.append(host_volume + ":" + self.container_volume_path)
            if self.container_user_id:
                docker_cmd.append("--user")
                docker_cmd.append(self.container_user_id)

            docker_cmd.append(self.container_image)

            if not self.cmd and not self.container_shell_path:
                fu.log(
                    "WARNING: The command-line is empty your container should know what to do automatically.",
                    self.out_log, self.global_log
                )
            else:
                cmd = ['"' + " ".join(self.cmd) + '"']
                docker_cmd.append(self.container_shell_path)
                docker_cmd.extend(cmd)
            self.cmd = docker_cmd
        # Pcocc
        elif self.container_path.endswith("pcocc"):
            # pcocc run -I racov56:pmx cli.py mutate -h
            fu.log(
                "Using pcocc image %s" % self.container_image,
                self.out_log,
                self.global_log,
            )
            pcocc_cmd = [
                self.container_path,
                self.container_generic_command,
                "-I",
                self.container_image,
            ]
            if self.container_working_dir:
                pcocc_cmd.append("--cwd")
                pcocc_cmd.append(self.container_working_dir)
            if self.container_volume_path:
                pcocc_cmd.append("--mount")
                pcocc_cmd.append(host_volume + ":" + self.container_volume_path)
            if self.container_user_id:
                pcocc_cmd.append("--user")
                pcocc_cmd.append(self.container_user_id)

            if not self.cmd and not self.container_shell_path:
                fu.log(
                    "WARNING: The command-line is empty your container should know what to do automatically.",
                    self.out_log,
                    self.global_log,
                )
            else:
                cmd = ['\\"' + " ".join(self.cmd) + '\\"']
                pcocc_cmd.append(self.container_shell_path)
                pcocc_cmd.extend(cmd)
            self.cmd = pcocc_cmd
        # Local execution
        else:
            pass
            # fu.log('Not using any container', self.out_log, self.global_log)

    def execute_command(self):

        cwd = os.getcwd()
        if self.chdir_sandbox:
            os.chdir(self.stage_io_dict["unique_dir"])

        self.return_code = cmd_wrapper.CmdWrapper(
            cmd=self.cmd,
            shell_path=self.shell_path,
            out_log=self.out_log,
            err_log=self.err_log,
            global_log=self.global_log,
            env=self.env_vars_dict,
            timeout=self.timeout,
            disable_logs=self.disable_logs
        ).launch()

        if self.chdir_sandbox:
            os.chdir(cwd)

    def run_biobb(self):
        self.create_cmd_line()
        self.execute_command()

    def copy_to_host(self):
        """Copy output files from the sandbox to the host system."""
        for file_ref, file_path in self.stage_io_dict["out"].items():
            dest_path = Path(self.io_dict["out"][file_ref])

            # For directories, we need to ensure the directory exists in the sandbox
            if self.doc_arguments_dict.get(file_ref, {}).get('type') == 'dir':
                # If the output is a directory, ensure it exists in the sandbox
                sandbox_dir_path = Path(self.stage_io_dict["unique_dir"]).joinpath(file_path)
                fu.log(f"Copy directory to host: {sandbox_dir_path} --> {dest_path}", self.out_log, self.global_log)
                fu.copytree_new_files_only(sandbox_dir_path, dest_path)
            else:
                if not file_path:
                    continue
                sandbox_file_path = Path(self.stage_io_dict["unique_dir"]).joinpath(Path(file_path).name)
                # Ensure file exists in the sandbox
                if not sandbox_file_path.exists():
                    continue
                # Only copy if destination doesn't exist or is different from source
                if not dest_path.exists() or not sandbox_file_path.samefile(dest_path):
                    shutil.copy2(sandbox_file_path, dest_path)

    def create_tmp_file(self, extension: str) -> None:
        """Create a temporary file in the unique directory. These files are
        removed when self.remove_tmp_files is called."""
        tmp_file = fu.create_unique_file_path(self.stage_io_dict["unique_dir"], extension)
        self.tmp_files.append(tmp_file)
        return tmp_file

    def create_tmp_dir(self) -> None:
        """Create a temporary directory in the unique directory. These directories are
        removed when self.remove_tmp_files is called."""
        tmp_dir = fu.create_unique_dir(self.stage_io_dict["unique_dir"], "tmpdir_", self.out_log)
        self.tmp_files.append(tmp_dir)
        return tmp_dir

    def remove_tmp_files(self):
        # Make sure current directory is not in the tmp_files list
        if str(os.getcwd()) in self.tmp_files:
            self.tmp_files.remove(str(os.getcwd()))

        if self.remove_tmp:
            fu.rm_file_list(self.tmp_files, self.out_log)

    @classmethod
    def get_main(cls, launcher, description, custom_flags=None):
        """Get command line execution of this building block. Please check the command line documentation."""
        def main():
            # Get the arguments and properties from the class docstring
            doc_arguments_dict, _ = fu.get_doc_dicts(cls.__doc__)
            # Create the argument parser
            parser = argparse.ArgumentParser(description=description,
                                             formatter_class=lambda prog: argparse.RawTextHelpFormatter(prog, width=99999))
            parser.add_argument(
                "-c", "--config", required=False,
                help="This file can be a YAML file, JSON file or JSON string",
            )
            required_args = parser.add_argument_group("required arguments")
            optional_args = parser.add_argument_group("optional arguments")
            # Use the doc_arguments_dict to add arguments to the parser
            # If we have only one input or output argument, we can use shorthand flags -i/-o
            input_args = [arg for arg, arg_dict in doc_arguments_dict.items() if arg_dict.get("input_output", "").lower().startswith("input")]
            output_args = [arg for arg, arg_dict in doc_arguments_dict.items() if arg_dict.get("input_output", "").lower().startswith("output")]

            for argument, argument_dict in doc_arguments_dict.items():
                # Determine if we should add shorthand flags
                shorthand_flags = [f'--{argument}']

                # Check if custom flags are provided for this argument
                if custom_flags and argument in custom_flags:
                    shorthand_flags.insert(0, custom_flags[argument])
                elif len(input_args) == 1 and argument in input_args:
                    shorthand_flags.insert(0, '-i')
                elif len(output_args) == 1 and argument in output_args:
                    shorthand_flags.insert(0, '-o')
                help_str = argument_dict.get("description", "") + f". Accepted formats: {', '.join(argument_dict.get('formats', {}).keys())}."
                if argument_dict["optional"]:
                    optional_args.add_argument(*shorthand_flags, required=False, help=help_str)
                else:
                    required_args.add_argument(*shorthand_flags, required=True, help=help_str)
            # Parse the arguments from the command line
            args = parser.parse_args()
            args.config = args.config or "{}"
            # Get the properties from the configuration yaml
            properties = settings.ConfReader(config=args.config).get_prop_dic()
            args_dict = vars(args)
            args_dict.pop('config', None)
            # Remove keys with None values from args_dict
            args_dict = {k: v for k, v in args_dict.items() if v is not None}
            # Return the function without executing it
            launcher(**args_dict, properties=properties)
        return main



"""Boiler plate functions for testsys
"""
import os
import pickle
from typing import Optional, Union, Any
from pathlib import Path
import sys
import shutil
import hashlib
from Bio.PDB import Superimposer, PDBParser  # type: ignore
import codecs
from biobb_common.configuration import settings
from biobb_common.tools import file_utils as fu
import numpy as np
import json
import jsonschema


def test_setup(test_object, dict_key: Optional[str] = None, config: Optional[str] = None):
    """Add the unitest_dir, test_dir, conf_file_path, properties and path as
    attributes to the **test_object** and create a directory to launch the unitest.

    Args:
        test_object (:obj:`test`): The test object.
        dict_key (str): Key of the test parameters in the yaml config file.
        config (str): Path to the configuration file.
    """
    test_object.testfile_dir = str(Path(Path(str(sys.modules[test_object.__module__].__file__)).resolve()).parent)
    test_object.unitest_dir = str(Path(test_object.testfile_dir).parent)
    test_object.test_dir = str(Path(test_object.unitest_dir).parent)
    test_object.data_dir = str(Path(test_object.test_dir).joinpath('data'))
    test_object.reference_dir = str(Path(test_object.test_dir).joinpath('reference'))
    if config:
        test_object.conf_file_path = config
    else:
        test_object.conf_file_path = str(Path(test_object.test_dir).joinpath('conf.yml'))

    conf = settings.ConfReader(test_object.conf_file_path)

    if dict_key:
        test_object.properties = conf.get_prop_dic()[dict_key]
        test_object.paths = {k: v.replace('test_data_dir', test_object.data_dir, 1).replace('test_reference_dir', test_object.reference_dir, 1) for k, v in conf.get_paths_dic()[dict_key].items()}
    else:
        test_object.properties = conf.get_prop_dic()
        test_object.paths = {k: v.replace('test_data_dir', test_object.data_dir, 1).replace('test_reference_dir', test_object.reference_dir, 1) for k, v in conf.get_paths_dic().items()}

    fu.create_dir(test_object.properties['path'])
    os.chdir(test_object.properties['path'])


def test_teardown(test_object):
    """Remove the **test_object.properties['working_dir_path']**

    Args:
        test_object (:obj:`test`): The test object.
    """
    if test_object.properties.get('remove_tmp', True):
        unitests_path = Path(test_object.properties['path']).resolve().parent
        print(f"\nRemoving: {unitests_path}")
        shutil.rmtree(unitests_path)


def exe_success(return_code: int) -> bool:
    """Check if **return_code** is 0

    Args:
        return_code (int): Return code of a process.

    Returns:
        bool: True if return code is equal to 0
    """
    return return_code == 0


def not_empty(file_path: str) -> bool:
    """Check if file exists and is not empty.

    Args:
        file_path (str): Path to the file.

    Returns:
        bool: True if **file_path** exists and is not empty.
    """
    if file_path.endswith('.zip'):
        print("Checking if empty zip: "+file_path)
        dst = file_path[:-4] + '_unzipped'
        os.makedirs(dst, exist_ok=True)
        # Extract zip and get list of files
        unzipped_files = fu.unzip_list(file_path, dest_dir=dst)
        # Check if there are any files in the zip
        return len(unzipped_files) > 0
    elif Path(file_path).is_dir():
        print("Checking if empty dir: "+file_path)
        return len(os.listdir(file_path)) > 0

    print("Checking if empty file: "+file_path)
    return Path(file_path).is_file() and Path(file_path).stat().st_size > 0


def compare_hash(file_a: str, file_b: str) -> bool:
    """Compute and compare the hashes of two files"""
    print("Comparing: ")
    print("        File_A: "+file_a)
    print("        File_B: "+file_b)
    file_a_hash = hashlib.sha256(open(file_a, 'rb').read()).digest()
    file_b_hash = hashlib.sha256(open(file_b, 'rb').read()).digest()
    print("        File_A hash: "+str(file_a_hash))
    print("        File_B hash: "+str(file_b_hash))
    return file_a_hash == file_b_hash


def equal(file_a: str, file_b: str, ignore_list: Optional[list[Union[str, int]]] = None, **kwargs) -> bool:
    """Check if two files are equal"""
    if ignore_list:
        # Line by line comparison
        return compare_line_by_line(file_a, file_b, ignore_list)

    if file_a.endswith(".zip") and file_b.endswith(".zip"):
        return compare_zip(file_a, file_b)

    if file_a.endswith(".pdb") and file_b.endswith(".pdb"):
        return compare_pdb(file_a, file_b, **kwargs)

    if file_a.endswith(".top") and file_b.endswith(".top"):
        return compare_top_itp(file_a, file_b)

    if file_a.endswith(".itp") and file_b.endswith(".itp"):
        return compare_top_itp(file_a, file_b)

    if file_a.endswith(".gro") and file_b.endswith(".gro"):
        return compare_ignore_first(file_a, file_b)

    if file_a.endswith(".prmtop") and file_b.endswith(".prmtop"):
        return compare_ignore_first(file_a, file_b)

    if file_a.endswith(".inp") and file_b.endswith(".inp"):
        return compare_ignore_first(file_a, file_b)

    if file_a.endswith(".par") and file_b.endswith(".par"):
        return compare_ignore_first(file_a, file_b)

    if file_a.endswith((".nc", ".netcdf", ".xtc")) and file_b.endswith((".nc", ".netcdf", ".xtc")):
        return compare_size(file_a, file_b, kwargs.get('percent_tolerance', 1.0))

    if file_a.endswith(".xvg") and file_b.endswith(".xvg"):
        return compare_xvg(file_a, file_b, kwargs.get('percent_tolerance', 1.0))

    image_extensions = ('.png', '.jfif', '.ppm', '.tiff', '.jpg', '.dib', '.pgm', '.bmp', '.jpeg', '.pbm', '.jpe', '.apng', '.pnm', '.gif', '.tif')
    if file_a.endswith(image_extensions) and file_b.endswith(image_extensions):
        return compare_images(file_a, file_b, kwargs.get('percent_tolerance', 1.0))

    return compare_hash(file_a, file_b)


def compare_line_by_line(file_a: str, file_b: str, ignore_list: list[Union[str, int]]) -> bool:
    print(f"Comparing ignoring lines containing this words: {ignore_list}")
    print("     FILE_A: "+file_a)
    print("     FILE_B: "+file_b)
    with open(file_a) as fa, open(file_b) as fb:
        for index, (line_a, line_b) in enumerate(zip(fa, fb)):
            if index in ignore_list or any(word in line_a for word in ignore_list if isinstance(word, str)):
                continue
            elif line_a != line_b:
                return False
        return True


def equal_txt(file_a: str, file_b: str) -> bool:
    """Check if two text files are equal"""
    return compare_hash(file_a, file_b)


def compare_zip(zip_a: str, zip_b: str) -> bool:
    """ Compare zip files """
    print("This is a ZIP comparison!")
    print("Unzipping:")
    print("Creating a unique_dir for: %s" % zip_a)
    zip_a_dir = fu.create_unique_dir()
    zip_a_list = fu.unzip_list(zip_a, dest_dir=zip_a_dir)
    print("Creating a unique_dir for: %s" % zip_b)
    zip_b_dir = fu.create_unique_dir()
    zip_b_list = fu.unzip_list(zip_b, dest_dir=zip_b_dir)

    if not len(zip_a_list) == len(zip_b_list):
        return False

    for uncompressed_zip_a in zip_a_list:
        uncompressed_zip_b = str(Path(zip_b_dir).joinpath(Path(uncompressed_zip_a).name))
        if not equal(uncompressed_zip_a, uncompressed_zip_b):
            return False

    return True


def compare_pdb(pdb_a: str, pdb_b: str, rmsd_cutoff: int = 1, remove_hetatm: bool = True, remove_hydrogen: bool = True, **kwargs):
    """ Compare pdb files """
    print("Checking RMSD between:")
    print("     PDB_A: "+pdb_a)
    print("     PDB_B: "+pdb_b)
    pdb_parser = PDBParser(PERMISSIVE=True, QUIET=True)
    st_a = pdb_parser.get_structure("st_a", pdb_a)
    st_b = pdb_parser.get_structure("st_b", pdb_b)
    if st_a is None or st_b is None:
        print("    One of the PDB structures could not be parsed.")
        return False
    st_a = st_a[0]
    st_b = st_b[0]

    if remove_hetatm:
        print("     Ignoring HETAMT in RMSD")
        residues_a = [list(res.get_atoms()) for res in st_a.get_residues() if not res.id[0].startswith('H_')]
        residues_b = [list(res.get_atoms()) for res in st_b.get_residues() if not res.id[0].startswith('H_')]
        atoms_a = [atom for residue in residues_a for atom in residue]
        atoms_b = [atom for residue in residues_b for atom in residue]
    else:
        atoms_a = st_a.get_atoms()
        atoms_b = st_b.get_atoms()

    if remove_hydrogen:
        print("     Ignoring Hydrogen atoms in RMSD")
        atoms_a = [atom for atom in atoms_a if not atom.get_name().startswith('H')]
        atoms_b = [atom for atom in atoms_b if not atom.get_name().startswith('H')]

    atoms_a_list = list(atoms_a)
    atoms_b_list = list(atoms_b)
    print("     Atoms ALIGNED in PDB_A: "+str(len(atoms_a_list)))
    print("     Atoms ALIGNED in PDB_B: "+str(len(atoms_b_list)))
    super_imposer = Superimposer()
    super_imposer.set_atoms(atoms_a, atoms_b)
    super_imposer.apply(atoms_b)
    super_imposer_rms = super_imposer.rms if super_imposer.rms is not None else float('inf')
    print('     RMS: '+str(super_imposer_rms))
    print('     RMS_CUTOFF: '+str(rmsd_cutoff))
    return super_imposer_rms < rmsd_cutoff


def compare_top_itp(file_a: str, file_b: str) -> bool:
    """ Compare top/itp files """
    print("Comparing TOP/ITP:")
    print("     FILE_A: "+file_a)
    print("     FILE_B: "+file_b)
    with codecs.open(file_a, 'r', encoding='utf-8', errors='ignore') as f_a:
        next(f_a)
        with codecs.open(file_b, 'r', encoding='utf-8', errors='ignore') as f_b:
            next(f_b)
            return [line.strip() for line in f_a if not line.strip().startswith(';')] == [line.strip() for line in f_b if not line.strip().startswith(';')]


def compare_ignore_first(file_a: str, file_b: str) -> bool:
    """ Compare two files ignoring the first line """
    print("Comparing ignoring first line of both files:")
    print("     FILE_A: "+file_a)
    print("     FILE_B: "+file_b)
    with open(file_a) as f_a:
        next(f_a)
        with open(file_b) as f_b:
            next(f_b)
            return [line.strip() for line in f_a] == [line.strip() for line in f_b]


def compare_size(file_a: str, file_b: str, percent_tolerance: float = 1.0) -> bool:
    """ Compare two files using size """
    print("Comparing size of both files:")
    print(f"     FILE_A: {file_a}")
    print(f"     FILE_B: {file_b}")
    size_a = Path(file_a).stat().st_size
    size_b = Path(file_b).stat().st_size
    average_size = (size_a + size_b) / 2
    tolerance = average_size * percent_tolerance / 100
    tolerance_low = average_size - tolerance
    tolerance_high = average_size + tolerance
    print(f"     SIZE_A: {size_a} bytes")
    print(f"     SIZE_B: {size_b} bytes")
    print(f"     TOLERANCE: {percent_tolerance}%, Low: {tolerance_low} bytes, High: {tolerance_high} bytes")
    return (tolerance_low <= size_a <= tolerance_high) and (tolerance_low <= size_b <= tolerance_high)


def compare_xvg(file_a: str, file_b: str, percent_tolerance: float = 1.0) -> bool:
    """ Compare two files using size """
    print("Comparing size of both files:")
    print(f"     FILE_A: {file_a}")
    print(f"     FILE_B: {file_b}")
    arrays_tuple_a = np.loadtxt(file_a, comments=["@", '#'], unpack=True)
    arrays_tuple_b = np.loadtxt(file_b, comments=["@", '#'], unpack=True)
    for array_a, array_b in zip(arrays_tuple_a, arrays_tuple_b):
        if not np.allclose(array_a, array_b, rtol=percent_tolerance / 100):
            return False
    return True


def compare_images(file_a: str, file_b: str, percent_tolerance: float = 1.0) -> bool:
    try:
        from PIL import Image  # type: ignore
        import imagehash
    except ImportError:
        print("To compare images, please install the following packages: Pillow, imagehash")
        return False

    """ Compare two files using size """
    print("Comparing images of both files:")
    print(f"     IMAGE_A: {file_a}")
    print(f"     IMAGE_B: {file_b}")
    hash_a = imagehash.average_hash(Image.open(file_a))
    hash_b = imagehash.average_hash(Image.open(file_b))
    tolerance = (len(hash_a) + len(hash_b)) / 2 * percent_tolerance / 100
    if tolerance < 1:
        tolerance = 1
    difference = hash_a - hash_b
    print(f"     IMAGE_A HASH: {hash_a} SIZE: {len(hash_a)} bits")
    print(f"     IMAGE_B HASH: {hash_b} SIZE: {len(hash_b)} bits")
    print(f"     TOLERANCE: {percent_tolerance}%, ABS TOLERANCE: {tolerance} bits, DIFFERENCE: {difference} bits")
    if difference > tolerance:
        return False
    return True


def compare_object_pickle(python_object: Any, pickle_file_path: Union[str, Path], **kwargs) -> bool:
    """ Compare a python object with a pickle file """
    print(f"Loading pickle file: {pickle_file_path}")
    with open(pickle_file_path, 'rb') as f:
        pickle_object = pickle.load(f)

    # Special case for dictionaries
    if isinstance(python_object, dict) and isinstance(pickle_object, dict):
        differences = compare_dictionaries(python_object, pickle_object, ignore_keys=kwargs.get('ignore_keys', []), compare_values=kwargs.get('compare_values', True), ignore_substring=kwargs.get('ignore_substring', ""))
        if differences:
            print(50*'*')
            print("OBJECT:")
            print(python_object)
            print(50*'*')
            print()
            print(50*'*')
            print("EXPECTED OBJECT:")
            print(pickle_object)
            print(50*'*')

            print("Differences found:")
            for difference in differences:
                print(f"     {difference}")
            return False
        return True

    return python_object == pickle_object


def compare_dictionaries(dict1: dict, dict2: dict, path: str = "", ignore_keys: Optional[list[str]] = None, compare_values: bool = True, ignore_substring: str = "") -> list[str]:
    """Compare two dictionaries and print only the differences, ignoring specified keys."""
    if ignore_keys is None:
        ignore_keys = []

    differences = []

    # Get all keys from both dictionaries
    all_keys = set(dict1.keys()).union(set(dict2.keys()))

    for key in all_keys:
        if key in ignore_keys:
            continue
        if key not in dict1:
            differences.append(f"Key '{path + key}' found in dict2 but not in dict1")
        elif key not in dict2:
            differences.append(f"Key '{path + key}' found in dict1 but not in dict2")
        else:
            value1 = dict1[key]
            value2 = dict2[key]
            if isinstance(value1, dict) and isinstance(value2, dict):
                # Recursively compare nested dictionaries
                nested_differences = compare_dictionaries(value1, value2, path + key + ".", ignore_keys, compare_values, ignore_substring)
                differences.extend(nested_differences)
            elif (value1 != value2) and compare_values:
                if ignore_substring:
                    if (not str(value1).endswith(str(value2).replace(ignore_substring, ""))) and (not str(value2).endswith(str(value1).replace(ignore_substring, ""))):
                        differences.append(f"Difference at '{path + key}': dict1 has {value1}, dict2 has {value2}")

                else:
                    differences.append(f"Difference at '{path + key}': dict1 has {value1}, dict2 has {value2}")

    return differences


def validate_json(json_file_path: Union[str, Path], json_schema_path: Union[str, Path]) -> bool:
    """
    Validates a JSON file against a provided JSON schema.

    Args:
        json_file_path (str): Path to the JSON file to validate.
        json_schema_path (str): Path to the JSON schema file.

    Returns:
        bool: True if the JSON is valid, False if invalid.
    """
    print("Validating JSON file:")
    print(f"     JSON file: {json_file_path}")
    print(f"     JSON schema: {json_schema_path}")
    try:
        # Load the JSON file
        with open(json_file_path, 'r') as json_file:
            json_data = json.load(json_file)

        # Load the JSON schema
        with open(json_schema_path, 'r') as schema_file:
            schema = json.load(schema_file)

        # Validate the JSON data against the schema
        jsonschema.validate(instance=json_data, schema=schema)

        return True
    except jsonschema.ValidationError as ve:
        print(f"Validation error: {ve.message}")
        return False
    except json.JSONDecodeError as je:
        print(f"Invalid JSON format: {je.msg}")
        return False



#!/usr/bin/env python3

"""Module containing the haddock  class and the command line interface."""

import os
import shutil
from typing import Optional

from biobb_common.generic.biobb_object import BiobbObject
from biobb_common.tools import file_utils as fu
from biobb_common.tools.file_utils import launchlogger


class FolderTest(BiobbObject):
    """
    | biobb_haddock FolderTest
    | Wrapper class for the FolderTest module.
    | The FolderTest module.

    Args:
        input_folder (dir) (Optional): Path of the input folder. File type: input. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_haddock/master/biobb_haddock/test/reference/haddock/input_folder>`_. Accepted formats: directory (edam:format_1915).
        output_folder (dir): Path of the output folder. File type: output. `Sample file <https://raw.githubusercontent.com/bioexcel/biobb_haddock/master/biobb_haddock/test/reference/haddock/output_folder>`_. Accepted formats: directory (edam:format_1915).
        properties (dict - Python dictionary object containing the tool parameters, not input/output files):
            * **n** (*int*) - (4) Number of files create.
            * **file_prefix** (*str*) - ("file") Prefix for the created files.

    Examples:
        This is a use example of how to use the building block from Python::

            from biobb_haddock.haddock.folder_test import folder_test
            folder_test(input_folder='/path/to/input_folder',
                        output_folder='/path/to/output_folder',
                        properties={'n': 4})

    Info:
        * wrapped_software:
            * name: None
            * version: None
            * license: Apache-2.0
        * ontology:
            * name: EDAM
            * schema: http://edamontology.org/EDAM.owl
    """

    def __init__(
        self,
        output_folder: str,
        input_folder: Optional[str] = None,
        properties: Optional[dict] = None,
        **kwargs,
    ) -> None:
        properties = properties or {}

        # Call parent class constructor
        super().__init__(properties)
        self.locals_var_dict = locals().copy()

        # Input/Output files
        self.io_dict = {
            "in": {"input_folder": input_folder},
            "out": {"output_folder": output_folder},
        }

        self.n = properties.get('n', 4)
        self.file_prefix = properties.get('file_prefix', 'file')
        # Check the properties
        self.check_init(properties)

    @launchlogger
    def launch(self) -> int:
        """Execute the :class:`FolderTest <biobb_haddock.haddock.folder_test>` object."""
        # tmp_files = []

        # Setup Biobb
        if self.check_restart():
            return 0
        self.stage_files()

        sandbox_output_folder = self.stage_io_dict["out"]["output_folder"]
        os.makedirs(sandbox_output_folder, exist_ok=True)
        # Just move the input files to the output
        if self.stage_io_dict["in"].get("input_folder") and self.stage_io_dict["in"]["input_folder"] != sandbox_output_folder:
            shutil.copytree(self.stage_io_dict["in"]["input_folder"],
                            sandbox_output_folder+'/prev',
                            dirs_exist_ok=True)
        fu.log('Directory contents: ' + str(os.listdir(sandbox_output_folder)), self.out_log, self.global_log)
        # Create n files in the output folder
        fu.log(f"Creating {self.n} files in the output folder: {sandbox_output_folder}",
               self.out_log, self.global_log)
        for i in range(1, self.n + 1):
            with open(f'{sandbox_output_folder}/{self.file_prefix}_{i}.txt', 'w') as f:
                f.write(f"This is file number {i}")
        fu.log('Directory contents: ' + str(os.listdir(sandbox_output_folder)), self.out_log, self.global_log)
        # Copy files to host
        self.copy_to_host()

        # Remove temporal files
        self.remove_tmp_files()
        # self.check_arguments(output_files_created=True, raise_exception=False)
        return self.return_code


def folder_test(output_folder: str, input_folder: Optional[str] = None, properties: Optional[dict] = None, **kwargs) -> int:
    """Create :class:`FolderTest <biobb_haddock.haddock.folder_test>` class and
    execute the :meth:`launch() <biobb_haddock.haddock.folder_test.launch>` method."""
    return FolderTest(**dict(locals())).launch()


folder_test.__doc__ = FolderTest.__doc__
main = FolderTest.get_main(folder_test, "Wrapper of the HADDOCK3 FolderTest module.")

if __name__ == "__main__":
    main()



#!/usr/bin/env python3

"""Settings loader module.

This module contains the classes to read the different formats of the configuration files.

The configuration files are composed by paths to the files and properties. There are several common properties for all
the building blocks.

Some yaml files contain a tool key with the name of the tool to be executed inside the step key. The tool key is used
by the resp API to identify the tool to be executed.


Syntax:
    - **property** (*dataType*) - (Default value) Short description.

Available common step properties: (Each Biobb step also has their specific properties)
    - **tool** (*str*) - (None) Name of the tool to be executed, mostly used by the biobbAPI.
    - **global_log** (*Logger object*) - (None) Log from the main workflow.
    - **prefix** (*str*) - (None) Prefix if provided.
    - **step** (*str*) - (None) Name of the step.
    - **path** (*str*) - ('') Absolute path to the step working dir.
    - **working_dir_path** (*str*) - (Current working dir) Workflow output directory.
    - **global_properties_list** (*list*) - ([]) List of global properties.
"""

import yaml
import json
import logging
from pathlib import Path
from copy import deepcopy
from biobb_common.tools import file_utils as fu
from typing import Any, Optional

GALAXY_CHARACTER_MAP = {
    "__gt__": ">", "__lt__": "<", "__sq__": "'", "__dq__": '"', "__ob__": "[", "__cb__": "]",
    "__oc__": "{", "__cc__": "}", "__cn__": "\n", "__cr__": "\r", "__tc__": "\t", "__pd__": "#"}


def trans_galaxy_charmap(input_str):
    """Fixes escape characters introduced by Galaxy on Json inputs"""
    for ch in GALAXY_CHARACTER_MAP:
        input_str = input_str.replace(ch, GALAXY_CHARACTER_MAP[ch])
    return input_str


class ConfReader:
    """Configuration file loader for yaml format files.

    Args:
        config (str): Path to the configuration [YAML|JSON] file or JSON string.
        system (str): System name from the systems section in the configuration file.
    """

    def __init__(self, config: Optional[str] = None, *args, **kwargs):
        self.properties = self._read_config(config)
        self.global_properties = self._get_global_properties()
        self.working_dir_path = fu.get_working_dir_path(working_dir_path=self.global_properties.get("working_dir_path", None), restart=self.global_properties.get("restart", False))

    def get_working_dir_path(self) -> str:
        """get_working_dir_path() returns the working directory path.

        Returns:
            str: Working directory path.
        """
        return self.working_dir_path

    def _read_config(self, config: Optional[str] = None) -> dict[str, Any]:
        """_read_config() reads the configuration file and returns a dictionary.
        """
        if not config:
            return dict()
        config_dict = dict()
        config_tokens = str(config).split("#")
        if (json_string := config_tokens[0].strip()).startswith("{"):
            config_dict = json.loads(trans_galaxy_charmap(json_string))
        else:
            config_file_path = Path(config_tokens[0]).resolve()
            if not config_file_path.exists():
                raise FileNotFoundError(f"Configuration file {config_file_path} not found.")
            with open(config_file_path) as stream:
                try:
                    config_dict = yaml.safe_load(stream) or {}
                except yaml.YAMLError as yaml_error:
                    try:
                        config_dict = json.load(stream) or {}
                    except json.JSONDecodeError as json_error:
                        raise Exception(f"Error reading configuration file {config_file_path} is not a valid YAML: {yaml_error} or a valid JSON: {json_error}")

        # Read just one step specified in the configuration file path
        # i.e: Read just Editconf step from workflow_configuration.yaml file
        # "/home/user/workflow_configuration.yaml#Editconf"
        if len(config_tokens) > 1:
            return config_dict[config_tokens[1]]

        return config_dict

    def _get_global_properties(self) -> dict[str, Any]:
        """_get_global_properties() returns the global properties of the configuration file.

        Returns:
            dict: dictionary of global properties.
        """
        # Add default properties to the global properties
        return deepcopy((self.properties.get("global_properties") or {}))

    def _get_step_properties(self, key: str = "", prefix: str = "", global_log: Optional[logging.Logger] = None) -> dict[str, Any]:
        """_get_step_properties() returns the properties of the configuration file.

        Args:
            global_properties (dict): Global properties.
            key (str): Step name.
            prefix (str): Prefix if provided.
            global_log (Logger): Log from the main workflow.

        Returns:
            dict: dictionary of properties.
        """
        prop_dic = dict()
        prop_dic.update(deepcopy(self.global_properties))
        prop_dic["global_properties_list"] = list(self.global_properties.keys())
        prop_dic["step"] = key
        prop_dic["prefix"] = prefix
        prop_dic["global_log"] = global_log
        prop_dic["working_dir_path"] = self.working_dir_path
        prop_dic["path"] = str(Path(self.working_dir_path).joinpath(prefix, key))
        if key:
            prop_dic["tool"] = self.properties[key].get("tool", None)
            prop_dic.update(deepcopy((self.properties[key].get("properties") or {})))
        else:
            prop_dic["tool"] = self.properties.get("tool", None)
            prop_dic.update(deepcopy((self.properties.get("properties") or {})))

        return prop_dic

    def get_prop_dic(self, prefix: str = "", global_log: Optional[logging.Logger] = None) -> dict[str, Any]:
        """get_prop_dic() returns the properties dictionary where keys are the
        step names in the configuration YAML file and every value contains another
        nested dictionary containing the keys and values of each step properties section.
        All the paths in the system section are copied in each nested dictionary.
        For each nested dictionary the following keys are added:
            | **path** (*str*): Absolute path to the step working dir.
            | **step** (*str*): Name of the step.
            | **prefix** (*str*): Prefix if provided.
            | **global_log** (*Logger object*): Log from the main workflow.
            | **tool** (*str*): Name of the tool to be executed, mostly used by the biobbAPI.
            | **working_dir_path** (*str*): Workflow output directory.
            | **global_properties_list** (*list*): List of global properties.

        Args:
            prefix (str): Prefix if provided.
            global_log (:obj:Logger): Log from the main workflow.

        Returns:
            dict: dictionary of properties.
        """

        prop_dic: dict[str, Any] = dict()
        for key in self.properties:
            if key in ["global_properties", "paths", "properties", "tool"]:
                continue
            prop_dic[key] = self._get_step_properties(key=key, prefix=prefix, global_log=global_log)

        if not prop_dic:
            return self._get_step_properties(prefix=prefix, global_log=global_log)

        return prop_dic

    def get_paths_dic(self, prefix: str = "") -> dict[str, Any]:
        paths_dic: dict[str, Any] = dict()
        for key in self.properties:
            if key in ["global_properties", "paths", "properties", "tool"]:
                continue
            paths_dic[key] = self._get_step_paths(key=key, prefix=prefix)

        if not paths_dic:
            return self._get_step_paths(prefix=prefix)

        return paths_dic

    def _get_step_paths(self, key: str = "", prefix: str = "") -> dict[str, Any]:
        step_paths_dic = dict()
        if key:
            paths_dic = self.properties[key].get("paths", {})
        else:
            paths_dic = self.properties.get("paths", {})
        for file_key, path_value in paths_dic.items():
            if path_value.startswith("file:"):
                step_paths_dic[file_key] = path_value.replace("file:", "")
                continue
            step_paths_dic[file_key] = str(Path(self.working_dir_path).joinpath(prefix, self._solve_dependency(key, path_value)))

        return step_paths_dic

    def _solve_dependency(self, step, dependency_str: str) -> str:
        """_solve_dependency() solves the dependency of a path in the configuration file.
        """
        dependency_tokens = dependency_str.strip().split("/")
        if dependency_tokens[0] != "dependency":
            return str(Path(step).joinpath(dependency_str))

        if not step:
            raise Exception("Step name is required to solve dependency")

        return str(Path(dependency_tokens[1]).joinpath(self.properties.get(dependency_tokens[1], {}).get('paths', {}).get(dependency_tokens[2], "")))



import setuptools

with open("README.md", "r") as fh:
    long_description = fh.read()

setuptools.setup(
    name="biobb_common",
    version="5.2.1",
    author="Biobb developers",
    author_email="pau.andrio@bsc.es",
    description="Biobb_common is the base package required to use the biobb packages.",
    long_description=long_description,
    long_description_content_type="text/markdown",
    keywords="Bioinformatics Workflows BioExcel Compatibility",
    url="https://github.com/bioexcel/biobb_common",
    project_urls={
        "Documentation": "http://biobb-common.readthedocs.io/en/latest/",
        "Bioexcel": "https://bioexcel.eu/",
    },
    packages=setuptools.find_packages(exclude=["docs"]),
    package_data={'biobb_common': ['py.typed']},
    install_requires=["pyyaml", "requests", "biopython", "jsonschema"],
    python_requires='>=3.10',
    entry_points={
        "console_scripts": [
            "folder_test = biobb_common.generic.folder_test:main",
        ]
    },
    classifiers=[
        "Development Status :: 5 - Production/Stable",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "License :: OSI Approved :: Apache Software License",
        "Operating System :: MacOS :: MacOS X",
        "Operating System :: POSIX",
        "Operating System :: Unix"
    ],
)
